from __future__ import annotations
from . import exceptions
from .component import Component
from .mime_types import MIME_TYPES
import orm
import html
from . import get_geety_js


class Page:
    def __init__(self, components: dict, db_pools: list[orm.PoolProtocol], *, context: dict | None = None):
        self.components = components
        self.db_pools = db_pools
        self.entry_point = None
        self.styles = []
        self.scripts = []
        self.locale = 'ru_RU'
        self.title = 'Geety Page'
        self.favicon = None
        self.viewport = 'width=device-width, initial-scale=1'
        self.robots = ''
        self.description = ''
        self.keywords = ''
        self.author = ''
        self.canonical = ''
        self.image = ''
        self.type = 'website'
        self.card = 'summary'
        self.base_href = None
        self.base_target = '_self'
        self.links = []
        self.context = context or {}
        self.meta = [
            { 'charset': 'UTF-8' }
        ]
    
    def set_entry_point(self,
                        entry_point: str | Component) -> None:
        self.entry_point = self.components[entry_point] if issubclass(type(entry_point), str) else entry_point
    
    def add_style(self, style: str) -> None:
        self.styles.append(style)
    
    def add_script(self, script: str) -> None:
        self.scripts.append(script)
    
    def add_link(self, link_args: dict) -> None:
        self.links.append(link_args)
    
    def add_meta(
        self,
        *, 
        name: str | None = None,
        property: str | None = None,
        content: str | None = None) -> None:
        self.meta.append(
            ({ 'name': name } if name else {}) |
            ({ 'property': property } if property else {}) |
            ({ 'content': content } if content else {})
        )
    
    def set_title(self, title: str) -> None:
        self.title = title
    
    def set_favicon(self, favicon: str) -> None:
        self.favicon = favicon
    
    def set_base_href(self, base_href: str) -> None:
        self.base_href = base_href
    
    def set_base_target(self, base_target: str) -> None:
        self.base_target = base_target
    
    def set_viewport(self, viewport: str) -> None:
        self.viewport = viewport
    
    def set_robots(self, robots: str) -> None:
        self.robots = robots
    
    def set_description(self, description: str) -> None:
        self.description = description
    
    def set_keywords(self, keywords: str) -> None:
        self.keywords = keywords
    
    def set_author(self, author: str) -> None:
        self.author = author
    
    def set_canonical(self, canonical: str) -> None:
        self.canonical = canonical
    
    def set_image(self, image: str) -> None:
        self.image = image
    
    def set_type(self, _type: str) -> None:
        self.type = _type
    
    def set_locale(self, locale: str) -> None:
        self.locale = locale
    
    def set_card(self, card: str) -> None:
        self.card = card
                        
    async def html(self,
             *,
             component: Component | None = None,
             context: dict | None = None) -> str:
        if not self.entry_point:
            raise exceptions.EntryPointNotSet()

        body = await (component or self.entry_point).render(self, self.context|(context or {}))

        base = (
            '<base '
            f'target="{self.base_target}"'
            f'{'' if self.base_href is None else f'href="{self.base_href}"'}>'
        )

        if self.description:
            self.add_meta(name='description', content=self.description)
            self.add_meta(property='og:description', content=self.description)
            self.add_meta(name='twitter:description', content=self.description)
        if self.image:
            self.add_meta(property='og:image', content=self.image)
            self.add_meta(name='twitter:image', content=self.image)
        if self.type:
            self.add_meta(property='og:type', content=self.type)
        if self.card:
            self.add_meta(name='twitter:card', content=self.card)
        if self.favicon:
            self.add_link(
                { 'rel': 'icon', 'href': self.favicon } | 
                ({ 'type': MIME_TYPES[self.favicon.split('.')[-1]] }
                    if self.favicon.split('.')[-1] in MIME_TYPES else {})
            )
        if self.canonical:
            self.add_link({ 'rel': 'canonical', 'href': self.canonical })
        if self.keywords:
            self.add_meta(name='keywords', content=self.keywords)
        if self.author:
            self.add_meta(name='author', content=self.author)
        if self.viewport:
            self.add_meta(name='viewport', content=self.viewport)
        if self.robots:
            self.add_meta(name='robots', content=self.robots)

        meta = ''.join(['<meta ' + ' '.join([
            f'{key}="{html.escape(val)}"'
            for key, val in meta.items()
        ]) + '>' for meta in self.meta])

        links = ''.join(['<link ' + ' '.join([
            f'{key}="{html.escape(val)}" '
            for key, val in link.items()
        ]) + '>' for link in self.links])
    
        return (
            '<!DOCTYPE html>\n'
            f'<html {f'lang="{self.locale.split('_')[0]}"' if self.locale else ''}>'
                '<head>'
                    f'{meta}'
                    f'{base}'
                    f'{links}'
                    f'<title>{self.title}</title>'
                    f'<style>{''.join(''.join(self.styles).split('\n'))}</style>'
                    f'<script>{get_geety_js.GEETY_JS}</script>'
                '</head>'
                '<body>'
                    f'{body}'
                    f'<script>{'\n'.join(self.scripts)}</script>'
                '</body>'
            '</html>'
        )


from __future__ import annotations
from io import TextIOBase
from typing import Iterable
from . import exceptions
from .xml_parser import parse_component
from .page import Page
import orm


class App:
    def __init__(self, *, context: dict | None):
        self.db_pools = []
        self.components = {}
        self.context = context or {}
    
    @staticmethod
    def _load_file(file: TextIOBase, components_list: Iterable[str]) -> None:
        components = {}
        for child in parse_component(file).find_by_tag('Geety').children:
            if type(child) is str:
                continue
            if child.tag in components_list:
                raise exceptions.ComponentAlreadyExists(child.tag)
            child._is_def = True
            components[child.tag] = child
        return components
    
    def load(self, file: TextIOBase) -> None:
        self.components.update(App._load_file(file, self.components.keys()))
    
    def add_database_pool(self, pool: orm.PoolProtocol):
        self.db_pools.append(pool)

    def new_page(self, page_file):
        return Page(self.components | App._load_file(page_file, self.components.keys()), self.db_pools, context=self.context)



class GeetyException(Exception): ...


class EntryPointNotSet(GeetyException): 
    def __init__(self):
        super().__init__('entry point has not set')


class ComponentAlreadyExists(GeetyException):
    def __init__(self, tag):
        super().__init__(f'component {tag} has redefined twice')


class DBException(GeetyException): ...

class DBContextMismatch(DBException):
    def __init__(self, cont_index, op_index, op_name):
        super().__init__(f'Database #{op_index} operation {op_name} cannot be used inside Database #{cont_index} context')


class DBCollectionNotSpecified(DBException):
    def __init__(self, cont_index):
        super().__init__(f'collection in Database #{cont_index} has not specified')


from __future__ import annotations
from typing import Optional
from io import TextIOBase
from xml.etree import ElementTree as ET
from .component import Component


def parse_component(file: TextIOBase) -> Optional[Component]:
    # TODO: rewrite with ET.XMLPoolParser 
    def _elem_to_comp(elem: ET.Element, parent: Component | None = None) -> Component:
        comp = Component(tag=elem.tag, args=elem.attrib.copy(), parent=parent)
        if elem.text:
            comp.children.append(elem.text)
        for child in elem:
            comp.children.append(_elem_to_comp(child, parent=comp))
            if child.tail:
                comp.children.append(child.tail)
        return comp
    
    content = file.read()
    return _elem_to_comp(ET.fromstring(content)) if content else None


MIME_TYPES = {
    'txt': 'text/plain',
    'html': 'text/html',
    'htm': 'text/html',
    'css': 'text/css',
    'md': 'text/markdown',
    'markdown': 'text/markdown',
    'csv': 'text/csv',
    
    'js': 'application/javascript',
    'mjs': 'application/javascript',

    'json': 'application/json',
    'xml': 'application/xml',

    'bin': 'application/octet-stream',

    'bz2': 'application/x-bzip2',
    'gz': 'application/gzip',
    'jar': 'application/java-archive',
    'rar': 'application/vnd.rar',
    'tar': 'application/x-tar',
    'zip': 'application/zip',
    '7z': 'application/x-7z-compressed',

    'pdf': 'application/pdf',
    'rtf': 'application/rtf',

    'docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    'pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',

    'odt': 'application/vnd.oasis.opendocument.text',
    'ods': 'application/vnd.oasis.opendocument.spreadsheet',
    'odp': 'application/vnd.oasis.opendocument.presentation',

    'png': 'image/png',
    'jpg': 'image/jpeg',
    'jpeg': 'image/jpeg',
    'webp': 'image/webp',
    'svg': 'image/xvg+xml',
    'ico': 'image/x-icon',
    'bmp': 'image/bmp',
    'avif': 'image/avif',
    'tif': 'image/tiff',
    'tiff': 'image/tiff',

    'mp3': 'audio/mpeg',
    'wav': 'audio/wav',
    'weba': 'audio/webm',

    'mp4': 'video/mp4',
    'mpeg': 'video/mpeg',
    'webm': 'video/webm',

    'otf': 'font/otf',
    'ttf': 'font/ttf',
    'woff2': 'font/woff2',
    

    'aac': 'audio/aac',
    'abw': 'application/x-abiword',
    'apng': 'image/apng',
    'arc': 'application/x-freearc',
    'avi': 'video/x-msvideo',
    'azw': 'application/vnd.amazon.ebook',
    'bz': 'application/x-bzip',
    'cda': 'application/x-cdf',
    'csh': 'application/x-csh',
    'doc': 'application/msword',
    'eot': 'application/vnd.ms-fontobject',
    'ics': 'text/calendar',
    'jsonld': 'application/ld+json',
    'mid': 'audio/midi',
    'midi': 'audio/midi',
    'mpkg': 'application/vnd.apple.installer+xml',
    'oga': 'audio/ogg',
    'ogv': 'video/ogg',
    'ogx': 'application/ogg',
    'ogg': 'application/ogg',
    'opus': 'audio/ogg',
    'ppt': 'application/vnd.ms-powerpoint',
    'sh': 'application/x-sh',
    'ts': 'video/mp2t',
    'vsd': 'application/vnd.visio',
    'woff': 'font/woff',
    'xls': 'application/vnd.ms-excel',
    'xul': 'application/vnd.mozilla.xul+xml',
    '3gp': 'video/3gpp',
    '3g2': 'video/3gpp2'
}


from __future__ import annotations
from typing import Optional
from collections.abc import Generator
from itertools import chain
from urllib.parse import quote
from . import exceptions
from .property import Property
from .utils import list_strip, merge_dicts, get_nested_value, uid_generator
import html
import re
import orm
import json
import datetime  # required by eval


VAR_PATTERN = re.compile(r'\$([\w\.]+)(@html|@url)?')
INSTRUCTION_PATTERN = re.compile(r'^(\{(Geety|DB@[0-9]+|Fobox)\}[\w\.]+|style|link|base)$')
DB_QUERY_PATTERN = re.compile(r'^\{DB@[0-9]+\}Query$')
DB_OP_PATTERN = re.compile(r'^\{DB@([0-9]+)\}([\w\.]+)$')


async def var_pattern_apply_content(content, context, page, prevs):
    if not content:
        return content

    context = context.copy()
    pp_context = {}
    for var, _ in VAR_PATTERN.findall(content):
        if '.' in var and var not in context:
            context[var] = get_nested_value(context[var.split('.')[0]], '.'.join(var.split('.')[1:]))
        if issubclass(type(context.get(var)), Component):
            pp_context[var] = await context[var].render(page, context, prevs)
        else:
            pp_context[var] = str(context[var])

    def replacer(match):
        var = match.group(1)
        if match.group(2) == '@html':
            return html.escape(pp_context[var])
        elif match.group(2) == '@url':
            return quote(pp_context[var])
        else:
            return pp_context[var]
    
    return VAR_PATTERN.sub(replacer, content)


async def var_pattern_apply_args(args: dict, context: dict, page: 'Page', prevs: list[Component]) -> dict:
    return {
        key: await var_pattern_apply_content(val, context, page, prevs)
        for key, val in args.items()
    }

def eval_args(args: dict) -> dict:
    return {
        key: eval(val) if type(val) in (str, bytes) else val
        for key, val in args.items()
    }


async def cond_exec(condition: str, context: dict, page: 'Page', prevs: list[Component]):
    context = context.copy()
    new_context = {}
    for var, _ in VAR_PATTERN.findall(condition):
        if '.' in var:
            context[var] = get_nested_value(context[var.split('.')[0]], '.'.join(var.split('.')[1:]))
        if var not in context:
            return False
        if issubclass(type(context[var]), Component):
            new_context[var] = context[var].render(page, context, prevs)
        elif issubclass(type(context[var]), str):
            new_context[var] = '\'' + context[var] + '\''
        else:
            new_context[var] = context[var]
    return eval(await var_pattern_apply_content(condition, new_context, page, prevs))


class Component:
    def __init__(
            self,
            *,
            tag: str = '',
            args: dict | None = None,
            children: list[Component] | None = None,
            parent: Component | None = None,
            properties: list[Property] | None = None,
            uid: str | None = None,
            _is_def: bool = False):
        self.tag = tag
        self.args = args or {}
        self.children = list_strip(children or [])
        self.parent = parent
        self.properties = properties or []
        self._is_def = _is_def
        self.uid = uid or uid_generator()
            
    
    def copy(self) -> Component:
        return Component(
            tag=self.tag,
            args=self.args.copy(),
            children=[
                child if type(child) is str else child.copy()
                for child in self.children
            ],
            parent=self.parent,
            properties=[
                prop.copy()
                for prop in self.properties
            ],
            uid=self.uid,
            _is_def=self._is_def
        )
    
    async def render(self, page: 'Page', context: dict, prevs: list[Component] | None = None) -> str:
        def find_prev(prevs):
            for i in range(1, len(prevs) - 1):
                if prevs[-i].is_def():
                    return prevs[-i-1]
            return None
        
        prevs = prevs or []
        component = page.components.get(self.tag, self).copy()
        tag = component.args.get('{Geety}Extends', 'div') if self.tag in page.components else self.tag
        component.args.pop('{Geety}Extends', None)
        self.args.pop('{Geety}Extends', None)
        if self.tag in page.components and not self.is_def():
            context['GID'] = self.uid
        sargs = await var_pattern_apply_args(self.args, context, page, prevs+[self])
        if self.tag in page.components and not self.is_def():
            signature = eval_args(await var_pattern_apply_args({
                arg.args['name']: arg.args.get('def')
                for arg in component.find_all_by_tag('{Geety}Arg')
            }, context, page, prevs+[self]))
            if component.is_def():
                merge_dicts(signature, eval_args(await var_pattern_apply_args(self.args, context, page, prevs+[self])))
                merge_dicts(context, signature)
        cargs = await var_pattern_apply_args(component.args, context, page, prevs+[self])

        html = ''
        match tag:
            case '{Geety}Content':
                prev = find_prev(prevs)
                if type(prev) is RenderedComponent:
                    html += prev.html
            case '{Geety}Set':
                context.update(eval_args(sargs))
            case '{Geety}For':
                element_name, iterable_name = component.args['each'].split(':')
                if iterable_name == 'Content':
                    prev = find_prev(prevs)
                    if type(prev) is RenderedComponent:
                        html += prev.html
                else:
                    for elem in context[iterable_name]:
                        context[element_name] = elem
                        for subchild in component:
                            if type(subchild) is str:
                                html += await var_pattern_apply_content(subchild, context, page, prevs+[self])
                            else:
                                subchild = subchild.copy()
                                html += await subchild.render(page, context, prevs+[self])
            case '{Geety}If':
                if await cond_exec(self.args.get('cond', 'False'), context, page, prevs):
                    for child in self:
                        if type(child) is str:
                            html += await var_pattern_apply_content(child, context, page, prevs+[self])
                        else:
                            html += await child.render(page, context, prevs+[self])
            case '{Geety}Switch':
                for case in self.find_all_by_tag('{Geety}Case'):
                    if await cond_exec(case.args.get('cond', 'False'), context, page, prevs):
                        for child in case:
                            if type(child) is str:
                                html += await var_pattern_apply_content(child, context, page, prevs+[self])
                            else:
                                html += await child.render(page, context, prevs+[self])
            case '{Geety}Page':
                if 'title' in sargs:
                    page.set_title(sargs['title'])
                if 'description' in sargs:
                    page.set_description(sargs['description'])
                if 'image' in sargs:
                    page.set_image(sargs['image'])
                if 'type' in sargs:
                    page.set_type(sargs['type'])
                if 'card' in sargs:
                    page.set_card(sargs['card'])
                if 'favicon' in sargs:
                    page.set_favicon(sargs['favicon'])
                if 'canonical' in sargs:
                    page.set_canonical(sargs['canonical'])
                if 'locale' in sargs:
                    page.set_locale(sargs['locale'])
                if 'keywords' in sargs:
                    page.set_keywords(sargs['keywords'])
                if 'author' in sargs:
                    page.set_author(sargs['author'])
                if 'viewport' in sargs:
                    page.set_viewport(sargs['viewport'])
                if 'robots' in sargs:
                    page.set_robots(sargs['robots'])
            case 'style':
                style = ''
                for child in self:
                    if type(child) is str:
                        style += await var_pattern_apply_content(child, context, page, prevs+[self])
                    else:
                        style += await child.render(page, context, prevs+[self])
                page.add_style(style)
            case 'base':
                if 'href' in cargs:
                    page.set_base_href(cargs['href'])
                if 'target' in cargs:
                    page.set_base_target(cargs['target'])
            case 'link':
                page.add_link(cargs.copy())
            case tag if DB_QUERY_PATTERN.fullmatch(tag):
                cont_index = int(DB_OP_PATTERN.findall(tag)[0][0])
                query_id = self.args.get('id')
                async with await page.db_pools[cont_index-1].acquire() as conn:
                    if 'collection' not in self.args:
                        raise exceptions.DBCollectionNotSpecified(cont_index)
                    collection: orm.CollectionProtocol = getattr(conn, self.args['collection'])
                    result = []
                    if query_id is not None:
                        context[query_id] = result
                    for command in self:
                        if type(command) is str:
                            continue
                        op_index, op_name = DB_OP_PATTERN.findall(command.tag)[0]
                        op_index = int(op_index)
                        if cont_index != op_index:
                            raise exceptions.DBContextMismatch(cont_index, op_index, op_name)
                        cmd_args = await var_pattern_apply_args(command.args, context, page, prevs+[self])
                        _op_args = cmd_args.copy()
                        _op_args.pop('order', None)
                        _op_args.pop('fields', None)
                        _op_args.pop('returning', None)
                        op_args = eval_args(_op_args)
                        match op_name:
                            case 'Find':
                                result += await collection.find(
                                    op_args.get('filter', {}),
                                    fields=[
                                        field.strip() for field in 
                                        cargs.get('fields', '*').split(',')
                                    ],
                                    order=[
                                        orm.Order.ASC(o[1].strip()) if o[0].strip().lower() == 'asc' else orm.Order.DESC(o[1].strip())
                                        for o in [
                                            field.strip().split() for field in 
                                            cmd_args.get('order', '').split(',')
                                        ]
                                    ]
                                )
                            case 'Insert':
                                result.append(await collection.insert(
                                    op_args.get('values', {}),
                                    returning=[
                                        field.strip() for field in 
                                        cargs.get('returning', '').split(',')
                                    ]
                                ))
                            case 'Count':
                                result.append(await collection.count(
                                    op_args.get('filter', {})
                                ))
                            case 'Delete':
                                await collection.delete(
                                    op_args.get('filter', {})
                                )
                            case 'Update':
                                result += await collection.update(
                                    op_args.get('filter', {}),
                                    op_args.get('values', {}),
                                    returning=[
                                        field.strip() for field in 
                                        cargs.get('returning', '').split(',')
                                        if field.strip()
                                    ]
                                )
                            case 'Pop':
                                result += await collection.pop(
                                    op_args.get('filter', {}),
                                    returning=[
                                        field.strip() for field in 
                                        cargs.get('returning', '').split(',')
                                        if field.strip()
                                    ]
                                )
            case '{Geety}Arg': ...  # skip
            case '{Fobox}Properties':
                for prop in self.find_all_by_tag('{Fobox}Property'):
                    onload = prop.find_by_tag('{Fobox}PropertyOnLoad')
                    onchange = prop.find_by_tag('{Fobox}PropertyOnChange')
                    args = prop.find_by_tag('{Fobox}PropertyArgs')
                    pargs = {}
                    for arg in args.find_all_by_tag('{Fobox}PropertyArg') if args else ():
                        match arg.args.get('type'):
                            case 'value':
                                pargs[arg.args.get('name')] = await arg.render(page, context, prevs+[self, prop, args])
                            case 'map':
                                _map = {}
                                for value in arg.find_all_by_tag('{Fobox}PropertyValue'):
                                    _map[value.args.get('key')] = await value.render(page, context, prevs+[self, prop, args, arg])
                                pargs[arg.args.get('name')] = _map
                    prev = find_prev(prevs)
                    prev.properties.append(Property(
                        prop.args.get('type'), 
                        prop.args.get('key'),
                        prop.args.get('name'),
                        (await onload.render(page, context, prevs+[self, prop])).strip() if onload else None,
                        (await onchange.render(page, context, prevs+[self, prop])).strip() if onchange else None,
                        pargs
                    ))
                    page.add_script('try{'+f'({prev.properties[-1].onchange})(G.$(".{prevs[-2].uid}"), "{json.loads(prev.args.get('{Fobox}Properties', '{}')).get(prev.properties[-1].key, '')}")'+'}catch(e){console.warn("%cFailed to load Fobox Property\\n","font-size:x-large",e)}')
            case '{Fobox}PropertyOnLoad' | '{Fobox}PropertyOnChange' | '{Fobox}PropertyArg' | '{Fobox}PropertyValue':
                for child in self:
                    if type(child) is str:
                        html += await var_pattern_apply_content(child, context, page, prevs+[self])
                    else:
                        html += await child.render(page, context, prevs+[self])
            case '{Fobox}Name': ...  # skip
            case _:
                if self.tag in page.components and not self.is_def():
                    cargs['class'] = cargs.get('class', '') + ' ' + self.uid
                html += f'<{tag} {' '.join([f"{key}=\"{val}\"" for key, val in cargs.items()])}>' if not self.is_def() else ''
                if self.tag not in page.components or self.is_def():
                    for child in self:
                        if type(child) is str:
                            html += await var_pattern_apply_content(child.strip(), context, page, prevs+[self])
                        else:
                            html += await child.render(page, context, prevs+[self])
                elif self.tag in page.components and not self.is_def():
                    component.args = await var_pattern_apply_args(component.args, context, page, prevs+[self])
                    context.update(component.args)
                    html += await page.components[self.tag].copy().render(page, context, prevs+[RenderedComponent.from_component(
                        ''.join([
                                await var_pattern_apply_content(subchild, context, page, prevs+[self]) if type(subchild) is str else await subchild.render(page, context, prevs+[self])
                                for subchild in self
                        ]), self
                    )])
        html += f'</{tag}>' if not INSTRUCTION_PATTERN.findall(tag) and not self.is_def() else ''
        return html
    
    def query(self, query):
        # TODO: add all type of queries (for tags ' ', for classes '.', for ids '#', for children '>')
        spl = list(filter(str, re.split(r'([>\.# ])', query)))
        print(spl)
        if len(spl) % 2 == 1:
            spl = [' '] + spl
        it = iter(spl)
        subqueries = list(zip(it, it))
        match subqueries[0][0]:
            case ' ':
                component = self.find_by_tag(subqueries[0][1])
                if len(subqueries) > 1:
                    return component.query(''.join(chain.from_iterable(subqueries[1:])))
                return component
    
    def find_by_tag(self, tag: str) -> Optional[Component]:
        return next(self.find_all_by_tag(tag), None)

    def find_all_by_tag(self, tag: str) -> Generator[Component]:
        if self.tag == tag:
            yield self
        for child in self.children:
            if type(child) is not str:
                yield from child.find_all_by_tag(tag)
    
    def is_def(self):
        return self._is_def
                    
    def __repr__(self) -> str:
        return f'<Component {self.uid} {'definition ' if self.is_def() else ''}{self.tag} [{", ".join([f"{key}={val}" for key, val in self.args.items()])}] with {len(self.children)} children>'

    def __iter__(self) -> Generator[Component]:
        yield from self.children


class RenderedComponent(Component):
    def __init__(
            self,
            html: str,
            *,
            tag: str = '',
            args: dict | None = None,
            children: list[Component] | None = None,
            parent: Component | None = None,
            uid: str | None = None,
            _is_def: bool = False):
        Component.__init__(
            self,
            tag=tag,
            args=args,
            children=children,
            parent=parent,
            uid=uid,
            _is_def=_is_def
        )
        self.html = html

    def copy(self) -> Component:
        return Component(
            tag=self.tag,
            args=self.args.copy(),
            children=[
                child if type(child) is str else child.copy()
                for child in self.children
            ],
            parent=self.parent,
            uid=self.uid,
            _is_def=self._is_def
        )

    def render(self, components: dict, context: dict, prevs: list[Component], meta: dict) -> str:
        return self.html

    @staticmethod
    def from_component(html: str, component: Component) -> RenderedComponent:
        component = component.copy()
        return RenderedComponent(
            html,
            tag=component.tag,
            args=component.args,
            children=component.children,
            parent=component.parent,
            uid=component.uid,
            _is_def=component._is_def
        )
